package net.wanji.com.netty;


import cn.hutool.core.util.HexUtil;
import io.netty.bootstrap.Bootstrap;
import io.netty.buffer.Unpooled;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.DatagramPacket;
import io.netty.channel.socket.nio.NioDatagramChannel;
import lombok.extern.slf4j.Slf4j;
import net.wanji.com.cache.netty.NettyMessageCache;
import net.wanji.com.netty.handler.NettyServerHandler;
import net.wanji.com.pojo.netty.MessageResultPojo;
import org.apache.commons.lang3.StringUtils;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;

import java.net.InetSocketAddress;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

@Slf4j
@Component
public class NettyClient {

    private static final Map<String, ChannelId> IP_PORT_CHANNEL_ID_MAP = new HashMap<>();
    private static ChannelFuture udpChannelFuture = null;

    /**
     * 建立连接
     */
    public static void start(int port, ThreadPoolTaskExecutor threadPoolExecutor) throws InterruptedException {
        EventLoopGroup group = new NioEventLoopGroup();
        Bootstrap bootstrap = new Bootstrap();
        bootstrap.group(group)
                .channel(NioDatagramChannel.class)
                .option(ChannelOption.RCVBUF_ALLOCATOR, new FixedRecvByteBufAllocator(65535))
//                .localAddress(new InetSocketAddress(5051))
                // 绑定bending端口
                .remoteAddress(new InetSocketAddress(5051))
                .handler(new ChannelInitializer<NioDatagramChannel>() {
                    @Override
                    protected void initChannel(NioDatagramChannel datagramChannel) {
                        datagramChannel.pipeline()
                                .addLast(new NettyServerHandler(threadPoolExecutor));
                    }
                });
        ChannelFuture channelFuture = bootstrap.bind(5050).sync();
        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            group.shutdownGracefully();
            channelFuture.channel().closeFuture().syncUninterruptibly();
            log.warn("server is closed");
        }));
        log.info("udp application is running. binding port is {}", port);
        NettyClient.udpChannelFuture = channelFuture;
    }

    public static void sendMsg(String ip, Integer port, String msg) {
        byte[] bytes = HexUtil.decodeHex(msg);
        try {
            if (udpChannelFuture == null) {
                return;
            }
            udpChannelFuture.channel()
                    .writeAndFlush(new DatagramPacket(Unpooled.copiedBuffer(bytes), new InetSocketAddress(ip, port))).sync();
            log.info("send msg {} to {}/{}", msg, ip, port);
        } catch (InterruptedException e) {
            log.error("sendMsg is error", e);
        }
    }

    public static MessageResultPojo sendMsg(String ip, Integer port, String msg, String command, int timeout) {
        try {
            if (timeout > 0) {
                CountDownLatch countDownLatch = new CountDownLatch(1);
                String key = StringUtils.joinWith("#", ip, port, command);
                long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
                MessageResultPojo messageResultPojo = new MessageResultPojo(command, countDownLatch, now, timeout, null);
                NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.put(key, messageResultPojo);
                sendMsg(ip, port, msg);
                countDownLatch.await(300, TimeUnit.MILLISECONDS);
                return NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.remove(key);
            }
        } catch (InterruptedException e) {
            log.error("sendMsg is error", e);
        }
        return null;
    }

    @Scheduled(fixedRate = 300)
    public void checkWaitTimeout() {
        long now = LocalDateTime.now().toInstant(ZoneOffset.of("+8")).toEpochMilli();
        NettyMessageCache.NETTY_MESSAGE_RESULT_MAP.forEach((k, v) -> {
//            threadPoolExecutor.execute(() -> {
                if (now - v.getStartTime() > v.getWaitMillisecond()) {
                    log.warn("wait {}ms {} timeout", v.getWaitMillisecond(), k);
                    v.getCountDownLatch().countDown();
                }
            });
//        });
    }
}